import logging
import time
from functools import partial
from pathlib import Path
from typing import Optional, cast

from cryptography.fernet import Fernet
from selenium import webdriver
from selenium.common import TimeoutException
from selenium.webdriver import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support.ui import Select

from ibeam.src.handlers.secrets_handler import SecretsHandler
from ibeam.src.login.driver import DriverFactory, start_up_browser, save_screenshot, shut_down_browser
from ibeam.src.login.targets import Targets, targets_from_versions, is_present, Target, identify_target, find_element, has_text, is_visible, is_clickable
from ibeam.src.two_fa_handlers.notification_resend_handler import NotificationResendTwoFaHandler
from ibeam.src.two_fa_handlers.two_fa_handler import TwoFaHandler
from ibeam.src.utils.py_utils import exception_to_string
from ibeam.src.utils.selenium_utils import any_of

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)



class AttemptException(Exception):
    def __init__(self, *args, cause:str, **kwargs):
        self.cause = cause
        super().__init__(*args, **kwargs)

def check_version(driver) -> int:
    """ Check for the IBRK website version. Currently, there are various versions shown to users and we want to know which one we are operating on.

    Versions:

    * 1 = available until March 2023
    * 2 = available from March 2023
    """
    try:
        user_name_present = EC.presence_of_element_located((By.NAME, 'user_name'))
        WebDriverWait(driver, 5).until(user_name_present)
        return 1
    except TimeoutException as e:
        pass

    try:
        user_name_present = EC.presence_of_element_located((By.NAME, 'username'))
        WebDriverWait(driver, 5).until(user_name_present)
        return 2
    except TimeoutException as e:
        pass

    _LOGGER.warning(f'Cannot determine the version of IBKR website, assuming version 1')

    return 1


def _wait_and_identify_trigger(targets: Targets,
                               driver: webdriver.Chrome,
                               timeout: int,
                               *expected_conditions,
                               skip_identify: bool = False,
                               ) -> (WebElement, Target):
    trigger = WebDriverWait(driver, timeout).until(any_of(*expected_conditions))

    if skip_identify:
        return trigger, None

    target = identify_target(trigger, targets)
    _LOGGER.debug(f'target: {target}')

    return trigger, target

def handle_two_fa(two_fa_handler:TwoFaHandler, driver:webdriver.Chrome, strict_two_fa_code:bool) -> Optional[str]:
    _LOGGER.info(f'Attempting to acquire 2FA code from: {two_fa_handler}')

    try:
        two_fa_code = two_fa_handler.get_two_fa_code(driver)
        if two_fa_code is not None:
            two_fa_code = str(two_fa_code)  # in case someone returns an integer
    except Exception as e:
        _LOGGER.error(f'Error encountered while acquiring 2FA code. \nException:\n{exception_to_string(e)}')
        return None

    _LOGGER.debug(f'2FA code returned: {two_fa_code}')

    if strict_two_fa_code and two_fa_code is not None and \
            (not two_fa_code.isdigit() or len(two_fa_code) != 6):
        _LOGGER.error(f'Illegal 2FA code returned: {two_fa_code}. Ensure the 2FA code contains 6 digits or disable this check by setting IBEAM_STRICT_TWO_FA_CODE to False.')
        return None

    return two_fa_code

class LoginHandler():
    _VERSIONS = {
        1: {
            'USER_NAME_EL': 'NAME@@user_name',
            'ERROR_EL': 'CSS_SELECTOR@@.alert.alert-danger.margin-top-10'
        },
        2: {
            'USER_NAME_EL': 'NAME@@username',
            'ERROR_EL': 'CSS_SELECTOR@@.xyz-errormessage'
        }
    }

    def __init__(self,
                 secrets_handler:SecretsHandler,
                 two_fa_handler:TwoFaHandler,
                 driver_factory:DriverFactory,
                 targets: Targets,
                 base_url: str,
                 route_auth: str,
                 two_fa_select_target: str,
                 strict_two_fa_code: bool,
                 max_immediate_attempts: int,
                 oauth_timeout: int,
                 max_presubmit_buffer: int,
                 min_presubmit_buffer: int,
                 max_failed_auth: int,
                 outputs_dir: str,
                 use_paper_account: bool = False,
                 ):

        self.secrets_handler = secrets_handler
        self.two_fa_handler = two_fa_handler
        self.driver_factory = driver_factory
        self.targets = targets

        self.base_url = base_url
        self.route_auth = route_auth
        self.two_fa_select_target = two_fa_select_target
        self.strict_two_fa_code = strict_two_fa_code
        self.max_immediate_attempts = max_immediate_attempts
        self.oauth_timeout = oauth_timeout
        self.max_presubmit_buffer = max_presubmit_buffer
        self.min_presubmit_buffer = min_presubmit_buffer
        self.max_failed_auth = max_failed_auth
        self.outputs_dir = outputs_dir
        self.use_paper_account = use_paper_account

        self.failed_attempts = 0
        self.presubmit_buffer = self.min_presubmit_buffer

    def step_login(self,
                   targets: Targets,
                   wait_and_identify_trigger: callable,
                   driver:webdriver.Chrome,
                   account: str,
                   password: str,
                   key: str,
                   presubmit_buffer: int,
                   ):

        if account is None:
            _LOGGER.error(f'Account cannot be None. Specify by setting IBEAM_ACCOUNT environment variable.')
            raise AttemptException(cause='shutdown')

        if password is None:
            _LOGGER.error(f'Password cannot be None. Specify by setting IBEAM_PASSWORD environment variable.')
            raise AttemptException(cause='shutdown')

        # input credentials
        user_name_el = find_element(targets['USER_NAME'], driver)
        password_el = find_element(targets['PASSWORD'], driver)

        wait_and_identify_trigger(is_clickable(targets['USER_NAME']))

        if self.use_paper_account:
            _LOGGER.info('Switching to paper mode')
            live_paper_toggle_el = find_element(targets['LIVE_PAPER_TOGGLE'], driver)
            live_paper_toggle_el.click()
            time.sleep(3)

        user_name_el.clear()
        password_el.clear()

        user_name_el.send_keys(account)

        if key is None:
            password_el.send_keys(password)
        else:
            password_el.send_keys(Fernet(key).decrypt(password.encode('utf-8')).decode("utf-8"))

        password_el.send_keys(Keys.TAB)

        # small buffer to prevent race-condition on client side
        time.sleep(presubmit_buffer)

        _LOGGER.info('Submitting the form')
        submit_form_el = find_element(targets['SUBMIT'], driver)
        submit_form_el.click()

        trigger, target = wait_and_identify_trigger(
            has_text(targets['SUCCESS']),
            is_visible(targets['TWO_FA']),
            is_visible(targets['TWO_FA_SELECT']),
            is_visible(targets['TWO_FA_NOTIFICATION']),
            is_visible(targets['ERROR']),
            is_clickable(targets['IBKEY_PROMO']),
        )

        return trigger, target

    def step_select_two_fa(self,
                           targets: Targets,
                           wait_and_identify_trigger: callable,
                           driver:webdriver.Chrome,
                           two_fa_select_target: str,
                           ):
        _LOGGER.info(f'Required to select a 2FA method.')
        select_el = find_element(targets['TWO_FA_SELECT'], driver)
        select = Select(select_el)
        select.select_by_visible_text(two_fa_select_target)

        trigger, target = wait_and_identify_trigger(
            has_text(targets['SUCCESS']),
            is_visible(targets['TWO_FA']),
            is_visible(targets['TWO_FA_NOTIFICATION']),
            is_visible(targets['ERROR']),
            is_clickable(targets['IBKEY_PROMO'])
        )

        _LOGGER.info(f'2FA method "{two_fa_select_target}" selected successfully.')
        return trigger, target


    def step_two_fa_notification(self,
                                 targets: Targets,
                                 wait_and_identify_trigger: callable,
                                 driver: webdriver.Chrome,
                                 two_fa_handler: TwoFaHandler,
                                 ):
        _LOGGER.info(f'Credentials correct, but Gateway requires notification two-factor authentication.')

        two_fa_handler = cast(two_fa_handler, NotificationResendTwoFaHandler)

        if two_fa_handler is not None:
            two_fa_success = two_fa_handler.interact_with_notification(driver, targets)
            if not two_fa_success:
                driver.refresh()
                raise AttemptException(cause='continue')

        trigger, target = wait_and_identify_trigger(
            has_text(targets['SUCCESS']),
            is_clickable(targets['IBKEY_PROMO']),
            is_visible(targets['ERROR'])
        )
        return trigger, target

    def step_two_fa(self,
                    targets: Targets,
                    wait_and_identify_trigger: callable,
                    driver: webdriver.Chrome,
                    two_fa_handler: TwoFaHandler,
                    strict_two_fa_code: bool,
                    ):
        _LOGGER.info(f'Credentials correct, but Gateway requires two-factor authentication.')
        if two_fa_handler is None:
            _LOGGER.critical(
                f'######## ATTENTION! ######## No 2FA handler found. You may define your own 2FA handler or use built-in handlers. See documentation for more: https://github.com/Voyz/ibeam/wiki/Two-Factor-Authentication')
            raise AttemptException(cause='shutdown')

        two_fa_code = handle_two_fa(two_fa_handler, driver, strict_two_fa_code)

        if two_fa_code is None:
            _LOGGER.warning(f'No 2FA code returned. Aborting authentication.')
            raise AttemptException(cause='break')
        else:
            two_fa_el, _ = wait_and_identify_trigger(is_clickable(targets['TWO_FA_INPUT']), skip_identify=True)

            two_fa_el.clear()
            two_fa_el.send_keys(two_fa_code)

            _LOGGER.info('Submitting the 2FA form')
            two_fa_el.send_keys(Keys.RETURN)

            trigger, target = wait_and_identify_trigger(
                has_text(targets['SUCCESS']),
                is_clickable(targets['IBKEY_PROMO']),
                is_visible(targets['ERROR'])
            )

            return trigger, target

    def step_handle_ib_key_promo(self,
                                 driver: webdriver.Chrome,
                                 targets: Targets,
                                 wait_and_identify_trigger: callable,
                                 ib_promo_key_trigger: WebElement,
                                 ):
        _LOGGER.info('Handling IB-Key promo display...')
        # ib_promo_key_trigger.click()
        time.sleep(3)
        driver.execute_script("arguments[0].click();", ib_promo_key_trigger)

        trigger, target = wait_and_identify_trigger(
            has_text(targets['SUCCESS']),
            is_visible(targets['ERROR'])
        )

        return trigger, target

    def step_paper_toggle(self,
                          driver:webdriver.Chrome,
                          targets: Targets,
                          wait_and_identify_trigger: callable
                          ):
        _LOGGER.info('Switching to paper mode and reattempting to submit the form')
        live_paper_toggle_el = find_element(targets['LIVE_PAPER_TOGGLE'], driver)
        live_paper_toggle_el.click()

        time.sleep(3)

        submit_form_el = find_element(targets['SUBMIT'], driver)
        submit_form_el.click()

        time.sleep(3)

        trigger, target = wait_and_identify_trigger(
            has_text(targets['SUCCESS']),
            is_visible(targets['TWO_FA']),
            is_visible(targets['TWO_FA_SELECT']),
            is_visible(targets['TWO_FA_NOTIFICATION']),
            is_visible(targets['ERROR']),
            is_clickable(targets['IBKEY_PROMO']),
        )

        return trigger, target

    def step_error(self,
                   driver:webdriver.Chrome,
                   error_trigger: WebElement,
                   max_presubmit_buffer: int,
                   max_failed_auth: int,
                   outputs_dir: str
                   ):
        _LOGGER.error(f'Error displayed by the login webpage: {error_trigger.text}')
        save_screenshot(driver, outputs_dir, '__failed_attempt')
        if error_trigger.text == 'Invalid username password combination' and self.presubmit_buffer < max_presubmit_buffer:
            # global _PRESUBMIT_BUFFER
            # _PRESUBMIT_BUFFER += 5
            self.presubmit_buffer += 5
            if self.presubmit_buffer >= self.max_presubmit_buffer:
                self.presubmit_buffer = self.max_presubmit_buffer
                _LOGGER.warning(f'The presubmit buffer set to maximum: {self.max_presubmit_buffer}')
            else:
                _LOGGER.warning(f'Increased presubmit buffer to {self.presubmit_buffer}')

        # try to prevent having the account locked-out
        if error_trigger.text == 'failed' or error_trigger.text == 'Invalid username password combination' and max_failed_auth > 0:
            self.failed_attempts += 1
            if self.failed_attempts >= self.max_failed_auth:
                _LOGGER.critical(
                    f'######## ATTENTION! ######## Maximum number of failed authentication attempts (IBEAM_MAX_FAILED_AUTH={self.max_failed_auth}) reached. IBeam will shut down to prevent an account lock-out. It is recommended you attempt to authenticate manually in order to reset the counter. Read the execution logs and report issues at https://github.com/Voyz/ibeam/issues')
                raise AttemptException(cause='shutdown')

        time.sleep(1)
        raise AttemptException(cause='continue')


    def handle_timeout_exception(self,
                                 e:Exception,
                                 targets: Targets,
                                 driver:webdriver.Chrome,
                                 website_version: int,
                                 route_auth: str,
                                 base_url: str,
                                 outputs_dir: str):
        page_loaded_correctly = True
        try:
            website_loaded = EC.presence_of_element_located((By.CLASS_NAME, 'login'))
            WebDriverWait(driver, 5).until(website_loaded)
        except TimeoutException as _:
            page_loaded_correctly = False

        if not page_loaded_correctly or website_version == -1:
            _LOGGER.error(f'Timeout reached when waiting for authentication. The website seems to not be loaded correctly. Consider increasing IBEAM_PAGE_LOAD_TIMEOUT. \nWebsite URL: {base_url + route_auth} \n \nException:\n{exception_to_string(e)}')
        else:
            _LOGGER.error(f'Timeout reached searching for website elements, but the website seems to be loaded correctly. It is possible the setup is incorrect. \nWebsite version: {website_version} \nDOM elements searched for: {targets}. \nException:\n{exception_to_string(e)}')

        save_screenshot(driver, outputs_dir, '__timeout-exception')

    def step_failed_two_fa(self, driver:webdriver.Chrome):
        # this means no two_fa_code was returned and trigger remained the same - ie. don't authenticate
        time.sleep(1)
        driver.refresh()
        raise AttemptException(cause='continue')
        # todo: retry authentication or resend code

    def step_success(self):
        _LOGGER.info('Webpage displayed "Client login succeeds"')
        self.failed_attempts = 0
        self.presubmit_buffer = self.min_presubmit_buffer
        raise AttemptException(cause='success')


    def attempt(
            self,
            targets: Targets,
            wait_and_identify_trigger: callable,
            driver: webdriver.Chrome
    ):
        trigger, target = self.step_login(targets, wait_and_identify_trigger, driver, self.secrets_handler.account, self.secrets_handler.password, self.secrets_handler.key, self.presubmit_buffer)

        if target == targets['ERROR'] and trigger.text == 'You have selected the Live Account Mode, but the specified user is a Paper Trading user. Please select the correct Login mode.':
            trigger, target = self.step_paper_toggle(driver, targets, wait_and_identify_trigger)

        if target == targets['TWO_FA_SELECT']:
            trigger, target = self.step_select_two_fa(targets, wait_and_identify_trigger, driver, self.two_fa_select_target)

        if target == targets['TWO_FA_NOTIFICATION']:
            trigger, target = self.step_two_fa_notification(targets, wait_and_identify_trigger, driver, self.two_fa_handler)

        if target == targets['TWO_FA']:
            trigger, target = self.step_two_fa(targets, wait_and_identify_trigger, driver, self.two_fa_handler, self.strict_two_fa_code)

        if target == targets['IBKEY_PROMO']:
            trigger, target = self.step_handle_ib_key_promo(driver, targets, wait_and_identify_trigger, trigger)

        if target == targets['ERROR']:
            self.step_error(driver, trigger, self.max_presubmit_buffer, self.max_failed_auth, self.outputs_dir)

        elif target == targets['TWO_FA']:
            self.step_failed_two_fa(driver)

        elif target == targets['SUCCESS']:
            self.step_success()

    def load_page(self, targets:Targets, driver:webdriver.Chrome, base_url: str, route_auth: str):
        driver.get(base_url + route_auth)

        website_version = check_version(driver)

        targets = targets_from_versions(targets, self._VERSIONS[website_version])
        _LOGGER.debug(f'Targets: {targets}')

        wait_and_identify_trigger = partial(_wait_and_identify_trigger, targets, driver, self.oauth_timeout)

        # wait for the page to load
        wait_and_identify_trigger(is_clickable(targets['USER_NAME']), skip_identify=True)
        _LOGGER.info('Gateway auth webpage loaded')

        return wait_and_identify_trigger

    def login(self) -> (bool, bool):
        """
        Logs into the currently running gateway.

        If both password and key are provided, cryptography.fernet decryption will be used.

        First boolean - whether authentication was successful
        Second boolean - whether max failed attempts was reached and IBeam should shut down

        :return: Whether authentication was successful and whether IBeam should shut down
        :rtype: (bool, bool)
        """
        display = None
        success = False
        driver = None
        website_version = -1
        targets = self.targets

        try:
            _LOGGER.info(f'Loading auth webpage at {self.base_url + self.route_auth}')
            driver, display = start_up_browser(self.driver_factory)

            wait_and_identify_trigger = self.load_page(targets, driver, self.base_url, self.route_auth)

            immediate_attempts = 0

            while immediate_attempts < max(self.max_immediate_attempts, 1):
                immediate_attempts += 1
                _LOGGER.info(f'Login attempt number {immediate_attempts}')

                try:
                    wait_and_identify_trigger(is_clickable(targets['USER_NAME']), skip_identify=True)
                except TimeoutException as e:
                    _LOGGER.info(f'Page loaded but {targets["USER_NAME"]} is not clickable. Reloading webpage.')
                    wait_and_identify_trigger = self.load_page(targets, driver, self.base_url, self.route_auth)

                try:
                    self.attempt(targets, wait_and_identify_trigger, driver)
                except AttemptException as e:
                    """ Not very proud of using exceptions for flow of control, but doing so simplifies the logic of splitting this while loop into functions and being able to break, continue, shutdown, etc. from within these functions."""

                    if e.cause == 'continue':
                        continue
                    elif e.cause =='success':
                        success = True
                        break
                    elif e.cause =='shutdown':
                        return False, True
                    elif e.cause == 'break':
                        break
                    else:
                        raise RuntimeError(f'Invalid AttemptException: {e}')


            time.sleep(1)
        except TimeoutException as e:
            self.handle_timeout_exception(e, targets, driver, website_version, self.route_auth, self.base_url, self.outputs_dir)
            success = False
        except Exception as e:
            _LOGGER.error(f'Error encountered during authentication \nException:\n{exception_to_string(e)}')
            save_screenshot(driver, self.outputs_dir, '__generic-exception')
            success = False
        finally:
            shut_down_browser(driver, display)

        return success, False



import datetime
import logging
import os
import sys
from pathlib import Path

initialized = False


def initialize(log_format: str,
               log_level: str,
               log_to_file: bool,
               outputs_dir: str
               ):
    global initialized
    if initialized: return
    initialized = True

    logger = logging.getLogger('ibeam')
    formatter = logging.Formatter(log_format)

    # stdout handler, for INFO and below:
    h1 = logging.StreamHandler(stream=sys.stdout)
    h1.setLevel(getattr(logging, log_level))
    h1.addFilter(lambda record: record.levelno <= logging.INFO)
    h1.setFormatter(formatter)
    logger.addHandler(h1)

    # stderr handler, for WARNING and above:
    h2 = logging.StreamHandler(stream=sys.stderr)
    h2.setLevel(logging.WARNING)
    h2.setFormatter(formatter)
    logger.addHandler(h2)


    logger.setLevel(logging.DEBUG)

    if log_to_file:
        file_handler = DailyRotatingFileHandler(os.path.join(outputs_dir, 'ibeam_log'))
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)


def set_level_for_all(logger, level):
    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)


class DailyRotatingFileHandler(logging.FileHandler):

    def __init__(self, *args, date_format='%Y-%m-%d', **kwargs):
        self.timestamp = None
        self.date_format = date_format
        super().__init__(*args, **kwargs)

    def get_timestamp(self):
        return datetime.datetime.now().strftime(self.date_format)

    def get_filename(self, timestamp):
        return f'{self.baseFilename}__{timestamp}.txt'

    def _open(self):
        self.timestamp = self.get_timestamp()
        try:
            return open(self.get_filename(self.timestamp), self.mode, encoding=self.encoding)
        except FileNotFoundError:
            Path(self.baseFilename).parent.mkdir(parents=True, exist_ok=True)
            return open(self.get_filename(self.timestamp), self.mode, encoding=self.encoding)

    def emit(self, record):
        if self.get_timestamp() != self.timestamp:
            self.stream = self._open()

        super().emit(record)



import logging
import time
from pathlib import Path

from ibeam.src.handlers.http_handler import Status, HttpHandler
from ibeam.src.handlers.login_handler import LoginHandler
from ibeam.src.handlers.process_handler import ProcessHandler

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)


def condition_authenticated_true(status:Status) -> bool:
    # unhappy cases, we don't need to keep on retrying
    if not status.running or not status.session:
        return True

    # happy case, wa don't need to keep on retrying
    if status.authenticated:
        return True

    return False

def condition_logged_out(status:Status) -> bool:
    # unhappy cases, we don't need to keep on retrying
    if not status.running or not status.session or status.competing:
        return True

    # happy case, wa don't need to keep on retrying
    if not status.connected and not status.authenticated:
        return True

    return False

def condition_not_competing(status:Status) -> bool:
    # unhappy cases, we don't need to keep on retrying
    if not status.running or not status.session:
        return True

    # happy case, wa don't need to keep on retrying
    if not status.competing and status.connected and status.authenticated:
        return True

    return False

class StrategyHandler():

    def __init__(self,
                 http_handler:HttpHandler,
                 login_handler:LoginHandler,
                 process_handler:ProcessHandler,
                 authentication_strategy:str,
                 reauthenticate_wait:int,
                 restart_failed_sessions:bool,
                 restart_wait:int,
                 max_reauthenticate_retries:int,
                 max_status_check_retries:int,
                 ):
        self.http_handler = http_handler
        self.login_handler = login_handler
        self.process_handler = process_handler
        self.authentication_strategy = authentication_strategy
        self.reauthenticate_wait = reauthenticate_wait
        self.restart_failed_sessions = restart_failed_sessions
        self.restart_wait = restart_wait
        self.max_reauthenticate_retries = max_reauthenticate_retries
        self.max_status_check_retries = max_status_check_retries

    def try_authenticating(self, request_retries=1) -> (bool, bool, Status):

        status = self.http_handler.get_status(max_attempts=request_retries)
        if status.authenticated and not status.competing:  # running, authenticated and not competing
            return True, False, status

        _LOGGER.info(str(status))

        if not status.running:  # no gateway running
            _LOGGER.error('Cannot communicate with the Gateway. Consider increasing IBEAM_GATEWAY_STARTUP')
            return False, False, status
        else:
            _LOGGER.info(f'Authentication strategy: "{self.authentication_strategy}"')

            if self.authentication_strategy == 'A':
                return self._authentication_strategy_A(status, request_retries)
            elif self.authentication_strategy == 'B':
                return self._authentication_strategy_B(status, request_retries)
            else:
                _LOGGER.error(f'Unknown authentication strategy: "{self.authentication_strategy}". Defaulting to strategy A.')
                return self._authentication_strategy_A(status, request_retries)

    def _authentication_strategy_A(self, status:Status, request_retries=1) -> (bool, bool, Status):
        if status.session:
            if not status.connected or status.competing:
                _LOGGER.info('Competing Gateway session found, restarting and logging in...')
                self._logout()

            _LOGGER.info('Gateway session found but not authenticated, logging in...')
        else:
            _LOGGER.info('No active sessions, logging in...')

        success, shutdown = self.login_handler.login()
        _LOGGER.info(f'Logging in {"succeeded" if success else "failed"}')
        if shutdown:
            return False, True, status
        if not success:
            return False, False, status

        time.sleep(3)  # buffer for session to be authenticated

        # double check if authenticated
        status = self.http_handler.get_status(max_attempts=max(request_retries, 2))
        if not status.authenticated:
            if status.session:
                _LOGGER.error('Logging in succeeded, but active session is still not authenticated')
                self.http_handler.reauthenticate()

                if self.reauthenticate_wait > 0:
                    _LOGGER.info(f'Waiting {self.reauthenticate_wait} seconds to reauthenticate before restarting.')
                    time.sleep(self.reauthenticate_wait)

                if self.restart_failed_sessions:
                    _LOGGER.info('Logging out and reattempting full authentication')
                    self._logout()
                    return self.try_authenticating(request_retries=request_retries)
            elif status.running:
                _LOGGER.error('Logging in succeeded but there are still no active sessions')
            else:
                _LOGGER.error('Logging in succeeded but now cannot communicate with the Gateway')
            return False, False, status
        elif not status.connected or status.competing:
            _LOGGER.info('Logging in succeeded, session is authenticated but competing, reauthenticating...')
            self.http_handler.reauthenticate()
            time.sleep(self.restart_wait)
            return False, False, status

        return True, False, status

    def _authentication_strategy_B(self, status:Status, request_retries=1) -> (bool, bool, Status):
        if not status.session:
            _LOGGER.info('No active sessions, logging in...')
            return self._log_in(status)

        elif not status.connected or status.competing:
            _LOGGER.info('Competing or disconnected Gateway session found, logging out and reauthenticating...')
            return self._reauthenticate(status, first_logout=True)
        else:
            _LOGGER.info('Active session found but not authenticated, reauthenticating...')
            return self._reauthenticate(status)

    def _log_in(self, status):
        try:
            success, shutdown = self.login_handler.login()
            _LOGGER.info(f'Logging in {"succeeded" if success else "failed"}')

            if not success or shutdown:
                return False, shutdown, status
        except Exception as e:
            _LOGGER.exception(f'Error logging in: {e}')
            return False, False, status

        return self._post_authentication()

    def _reauthenticate(self, status, first_logout=False):
        try:
            if first_logout:
                self._logout()
            self.http_handler.reauthenticate()
        except Exception as e:
            _LOGGER.exception(f'Error reauthenticating: {e}')
            return False, False, status

        return self._post_authentication()
    def _logout(self):
        try:
            logout_response = self.http_handler.logout()
            logout_success = logout_response.read().decode('utf8') == '{"status":true}'
            _LOGGER.info(f'Gateway logout {"successful" if logout_success else "unsuccessful"}')
        except Exception as e:
            _LOGGER.exception(f'Exception logging out: {e}')

    def _post_authentication(self):
        """This method double-checks that the authentication was successful, and if not, reauthenticates"""

        # if we only just logged in and succeeded, this will not reauthenticate but only check status
        status = self._repeatedly_reauthenticate(self.max_reauthenticate_retries, condition_authenticated_true)

        if not status.running or not status.session or not status.connected or status.competing or not status.authenticated:
            _LOGGER.error(f'Repeatedly reauthenticating failed {self.max_reauthenticate_retries} times. Killing the Gateway and restarting the authentication process.')

            try:
                success = self.process_handler.kill_gateway()
            except Exception as e:
                _LOGGER.exception(f'Error killing the Gateway: {e}')
                success = False

            if not success:
                _LOGGER.error(f'Killing the Gateway process failed')

            return False, False, status

        self.http_handler.initialise()

        return True, False, status


    def _repeatedly_check_status(self, max_attempts=1, condition:callable=condition_authenticated_true):
        if not callable(condition):
            raise ValueError(f'Condition must be a callable, found: "{type(condition)}": {condition}')

        status = None

        for attempt in range(max_attempts):
            status = self.http_handler.get_status()

            if condition(status):
                return status

            if attempt < max_attempts - 1:
                time.sleep(1)
                if attempt == 0:
                    _LOGGER.info(f'Repeating status check attempts another {max_attempts - attempt - 1} times')

        _LOGGER.info(f'Max status check retries reached after {max_attempts} attempts. Consider increasing the retries by setting IBEAM_MAX_STATUS_CHECK_ATTEMPTS environment variable')
        return status

    def _repeatedly_reauthenticate(self, max_attempts=1, condition:callable=condition_authenticated_true):
        if not callable(condition):
            raise ValueError(f'Condition must be a callable, found: "{type(condition)}": {condition}')

        status = None

        for attempt in range(max_attempts):
            status = self._repeatedly_check_status(self.max_status_check_retries, condition)
            _LOGGER.info(str(status))

            if condition(status):
                return status

            if attempt < max_attempts - 1:
                self.http_handler.reauthenticate()
                _LOGGER.info(f'Repeated reauthentication attempt number {attempt + 2}')

        _LOGGER.info(f'Max reauthenticate retries reached after {max_attempts} attempts. Consider increasing the retries by setting IBEAM_MAX_REAUTHENTICATE_RETRIES environment variable')
        return status


from ibeam.src.var import UNDEFINED


class Config():
    def __init__(self, env_variables):
        self._all_variables = env_variables.copy()
        self.manual_input_variables = self._all_variables.get('IBEAM_MANUAL_INPUT_VARIABLES', [])

    def __getattr__(self, key):
        try:
            value = self._all_variables[key]
            if value is UNDEFINED:
                raise ValueError(f'Variable "IBEAM_{key}" is undefined. Fix by setting "IBEAM_{key}" environment variable.')
            return value
        except KeyError:
            # we allow the user to manually input some variables
            if key in self.manual_input_variables:
                return input(f'Enter value for {key}: ')

            raise AttributeError(f'{key} is not a valid config key. Existing keys: {list(self._all_variables.keys())}')

    @property
    def all_variables(self):
        return self._all_variables.copy()


import base64
import logging
import os
from pathlib import Path
from typing import Optional
import requests


_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)

SECRETS_SOURCE_ENV = 'env'
SECRETS_SOURCE_FS = 'fs'
SECRETS_SOURCE_GCP_SECRETS = 'gcp_secrets_manager'

"""If IBEAM_SECRETS_SOURCE is set to
SECRETS_SOURCE_ENV, or if it is not set, then
environment values will be assumed to hold the
secret values directly.

If IBEAM_SECRETS_SOURCE is set to SECRETS_SOURCE_FS
then the environment values are assumed to be file
paths to read for the secret value.

If IBEAM_SECRETS_SOURCE is set to SECRETS_SOURCE_GCP_SECRETS 
then the environment values are assumed to be Google Cloud 
Platform secrets in form of:

[SECRET_NAME]/versions/[SECRET_VERSION] 

In this case IBEAM_GCP_BASE_URL must also be provided in form of:

https://secretmanager.googleapis.com/v1/projects/[PROJECT_ID]/secrets
"""

class SecretsHandler():
    ...
    def __init__(self,
                 secrets_source: str,
                 gcp_base_url: Optional[str] = None,
                 ):
        self.secrets_source = secrets_source
        self.gcp_base_url = gcp_base_url

        """Character encoding for secret files"""
        self.encoding = os.environ.get(
            'IBEAM_ENCODING', default='UTF-8')

    def secret_value(self, encoding, name: str,
                     lstrip=None, rstrip='\r\n') -> Optional[str]:
        """
        secret_value reads secrets from os.environ or from
        the filesystem.

        Given a name, such as 'IBEAM_ACCOUNT', it will
        examine os.environ for a value associated with that
        name.

        If no value has been set, None is returned.
        Otherwise the self.secrets_source will be evaluated
        to determine how to handle the value.

        If self.secrets_source has been set to
        SECRETS_SOURCE_ENV the os.environ value will be
        returned as the secret value.

        If self.secrets_source has been set to
        SECRETS_SOURCE_FS then the os.environ value is
        treated as a filesystem path.  The file will be read
        as text and its contents returned as the secret
        value.

        Parameters:
          name:
            The identifier for the value, e.g.,
            'IBEAM_ACCOUNT', 'IBEAM_PASSWORD', or
            'IBEAM_KEY'.
          lstrip:
            If not None, strip these characters from the
            left of the returned value (default: None).
          rstrip:
            If not None, strip these characters from the
            right of the returned value (default: '\r\n')
        Returns:
          If the name is not defined in os.environ then None
          is returned.

          If self.secrets_source is SECRETS_SOURCE_ENV then
          the os.environ value is returned as the secret
          value.

          If self.secrets_source is SECRETS_SOURCE_FS then
          the os.environ value is treated as file path.  The
          file is read as a text file and its contents
          returned as the secret value.

          If an error is encountered reading the file then
          an error is logged and None is returned.
        """
        # read the environment value for name
        value = os.environ.get(name)
        if value is None:
            # no key for this name, nothing to do
            return None

        if self.secrets_source == SECRETS_SOURCE_ENV:
            # treat environment values as the secrets themselves
            if lstrip is not None:
                value = value.lstrip(lstrip)

            if rstrip is not None:
                value = value.rstrip(rstrip)

            return value
        elif self.secrets_source == SECRETS_SOURCE_FS:
            # treat environment values as filesystem paths to the secrets
            if not os.path.isfile(value):
                _LOGGER.error(
                    f'Unable to read env value for {name}: value is not a file')
                return None

            try:
                with open(value, mode='rt', encoding=encoding) as fh:
                    secret = fh.read()

                    if lstrip is not None:
                        secret = secret.lstrip(lstrip)

                    if rstrip is not None:
                        secret = secret.rstrip(rstrip)

                    return secret
            except IOError:
                _LOGGER.error(
                    f'Unable to read env value for {name} as a file.')
                return None
        elif self.secrets_source == SECRETS_SOURCE_GCP_SECRETS:
            # get authentication token from GCP
            response = requests.get('http://169.254.169.254/computeMetadata/v1/instance/service-accounts/default/token', headers={'Metadata-Flavor': 'Google'})
            if response.status_code != 200:
                _LOGGER.error(f'Google Metadata request returned status code {response.status_code} :: {response.reason} :: {response.text}')
                return None

            access_token = response.json()['access_token']

            # get secret from GCP
            secret_url = self.gcp_base_url + '/' + value + ':access'
            response2 = requests.get(secret_url, headers={'authorization': f'Bearer {access_token}'})
            if response2.status_code!= 200:
                _LOGGER.error(f'Google Secret Manager request returned status code {response2.status_code} :: {response2.reason} :: {response2.text}')
                return None

            payload_data = response2.json()['payload']['data']
            try:
                base64_data =  base64.b64decode(payload_data)
                decoded_data = base64_data.decode('utf-8')
            except Exception as e:
                _LOGGER.error(f'Unable to decode secret value for {name}: {e}')
                return None
            return decoded_data

        else:
            _LOGGER.error(
                f'Unknown Secrets Source: {self.secrets_source}')
            return None


    @property
    def account(self):
        """IBKR account name."""
        return self.secret_value(self.encoding, 'IBEAM_ACCOUNT')

    @property
    def password(self):
        """IBKR account password."""
        return self.secret_value(self.encoding, 'IBEAM_PASSWORD')

    @property
    def key(self):
        """Key to the IBKR password."""
        return self.secret_value(self.encoding, 'IBEAM_KEY')


import traceback


# from https://stackoverflow.com/a/37135014/3508719
def exception_to_string(excp):
    stack = traceback.extract_stack()[:-2] + traceback.extract_tb(excp.__traceback__)
    pretty = traceback.format_list(stack)
    return ''.join(pretty) + '\n  {} {}'.format(excp.__class__,excp)


# from: https://github.com/drgarcia1986/simple-settings/pull/281/files
_MAP = {
    'y': True,
    'yes': True,
    't': True,
    'true': True,
    'on': True,
    '1': True,
    'n': False,
    'no': False,
    'f': False,
    'false': False,
    'off': False,
    '0': False
}

def strtobool(value):
    try:
        return _MAP[str(value).lower()]
    except KeyError:
        raise ValueError('"{}" is not a valid bool value'.format(value))


import json
import logging
import os
from pathlib import Path
from typing import Optional

import requests

from ibeam.src.two_fa_handlers.two_fa_handler import TwoFaHandler

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)

_EXTERNAL_REQUEST_METHOD = os.environ.get('IBEAM_EXTERNAL_REQUEST_METHOD', 'GET')
"""Method to use by the external request 2FA handler."""

_EXTERNAL_REQUEST_URL = os.environ.get('IBEAM_EXTERNAL_REQUEST_URL')
"""URL to use by the external request 2FA handler."""

_EXTERNAL_REQUEST_TIMEOUT = int(os.environ.get('IBEAM_EXTERNAL_REQUEST_TIMEOUT', 300))
"""Timeout for the external 2FA request."""

_EXTERNAL_REQUEST_PARAMS = os.environ.get('IBEAM_EXTERNAL_REQUEST_PARAMS')
"""JSON-formatted URL params to use by the external request 2FA handler."""

_EXTERNAL_REQUEST_DATA = os.environ.get('IBEAM_EXTERNAL_REQUEST_DATA')
"""JSON-formatted POST data to use by the external request 2FA handler."""

_EXTERNAL_REQUEST_HEADERS = os.environ.get('IBEAM_EXTERNAL_REQUEST_HEADERS')
"""JSON-formatted headers to use by the external request 2FA handler."""


def parse_json(s):
    if s is None:
        return None
    try:
        return json.loads(s)
    except Exception as e:
        _LOGGER.exception(f'Error loading JSON string: "{e}" | for JSON: {s}')
        return None


class ExternalRequestTwoFaHandler(TwoFaHandler):

    def __init__(self, method: str = None,
                 url: str = None,
                 timeout: int = None,
                 params=None,
                 data=None,
                 headers=None,
                 *args, **kwargs
                 ):

        self.method = method if method is not None else _EXTERNAL_REQUEST_METHOD
        self.url = url if url is not None else _EXTERNAL_REQUEST_URL
        self.timeout = timeout if timeout is not None else _EXTERNAL_REQUEST_TIMEOUT
        self.params = params if params is not None else parse_json(_EXTERNAL_REQUEST_PARAMS)
        self.data = data if data is not None else parse_json(_EXTERNAL_REQUEST_DATA)
        self.headers = headers if headers is not None else parse_json(_EXTERNAL_REQUEST_HEADERS)
        super().__init__(*args, **kwargs)

    def get_two_fa_code(self, _) -> Optional[str]:
        try:
            response = requests.request(method=self.method,
                                        url=self.url,
                                        timeout=self.timeout,
                                        params=self.params,
                                        data=self.data,
                                        headers=self.headers, )
            response.raise_for_status()
            return response.content.decode("utf-8")
        except requests.exceptions.HTTPError as err:
            _LOGGER.error(err)
            return None

    def __str__(self):
        return f"ExternalRequestTwoFaHandler(method={self.method}, url={self.url}, timeout={self.timeout}, params={self.params}, headers={self.headers})"



import argparse
import logging
import os
import signal
import sys
from pathlib import Path

def add_to_path():
    _this_filedir = os.path.abspath(os.path.dirname(__file__))
    sys.path.insert(0, str(Path(_this_filedir).parent))

add_to_path()

from ibeam.config import Config
from ibeam.src.handlers.login_handler import LoginHandler
from ibeam.src.handlers.process_handler import ProcessHandler
from ibeam.src.handlers.secrets_handler import SecretsHandler
from ibeam.src.handlers.strategy_handler import StrategyHandler
from ibeam.src.login.driver import DriverFactory, shut_down_browser
from ibeam.src.login.targets import create_targets

import ibeam

from ibeam.src.gateway_client import GatewayClient
from ibeam.src.handlers.http_handler import HttpHandler
from ibeam.src import var, two_fa_selector
from ibeam.src.handlers.inputs_handler import InputsHandler

_LOGGER = logging.getLogger('ibeam')

def parse_args():
    parser = argparse.ArgumentParser(description='Start, authenticate and verify the IB Gateway.')
    parser.add_argument('-a', '--authenticate', action='store_true', help='Authenticates the existing gateway.')
    parser.add_argument('-k', '--kill', action='store_true', help='Kill the gateway.')
    parser.add_argument('-m', '--maintain', action='store_true', help='Maintain the gateway.')
    parser.add_argument('-s', '--start', action='store_true', help='Start the gateway.')
    parser.add_argument('-t', '--tickle', action='store_true', help='Tickle the gateway.')
    parser.add_argument('-u', '--user', action='store_true', help='Get the user.')
    parser.add_argument('-c', '--check', action='store_true', help='Check if session is authenticated.')
    parser.add_argument('-v', '--verbose', action='store_true', help='Verbose output.')

    args = parser.parse_args()
    return args


if __name__ == '__main__':
    cnf = Config(var.all_variables)

    from ibeam.src import logs
    logs.initialize(
        log_format=cnf.LOG_FORMAT,
        log_level=cnf.LOG_LEVEL,
        log_to_file=cnf.LOG_TO_FILE,
        outputs_dir=cnf.OUTPUTS_DIR,
    )


    _LOGGER.info(f'############ Starting IBeam version {ibeam.__version__} ############')
    args = parse_args()

    if args.verbose:
        logs.set_level_for_all(_LOGGER, logging.DEBUG)

    inputs_handler = InputsHandler(inputs_dir=cnf.INPUTS_DIR, gateway_dir=cnf.GATEWAY_DIR)

    http_handler = HttpHandler(
        inputs_handler=inputs_handler,
        base_url=cnf.GATEWAY_BASE_URL,
        route_validate=cnf.ROUTE_VALIDATE,
        route_tickle=cnf.ROUTE_TICKLE,
        route_logout=cnf.ROUTE_LOGOUT,
        route_reauthenticate=cnf.ROUTE_REAUTHENTICATE,
        route_initialise=cnf.ROUTE_INITIALISE,
        request_timeout=cnf.REQUEST_TIMEOUT,
    )

    driver_factory = DriverFactory(
        driver_path=cnf.CHROME_DRIVER_PATH,
        ui_scaling=cnf.UI_SCALING,
        page_load_timeout=cnf.PAGE_LOAD_TIMEOUT,
    )

    two_fa_handler = two_fa_selector.select(
        handler_name=cnf.TWO_FA_HANDLER,
        driver_factory=driver_factory,
        outputs_dir=cnf.OUTPUTS_DIR,
        custom_two_fa_handler=cnf.CUSTOM_TWO_FA_HANDLER,
        inputs_dir=cnf.INPUTS_DIR,
    )

    _LOGGER.info(f'Secrets source: {cnf.SECRETS_SOURCE}')
    secrets_handler = SecretsHandler(secrets_source=cnf.SECRETS_SOURCE, gcp_base_url=cnf.GCP_SECRETS_URL)

    targets = create_targets(cnf)

    login_handler = LoginHandler(
        secrets_handler=secrets_handler,
        two_fa_handler=two_fa_handler,
        driver_factory=driver_factory,
        targets=targets,
        base_url=cnf.GATEWAY_BASE_URL,
        route_auth=cnf.ROUTE_AUTH,
        two_fa_select_target=cnf.TWO_FA_SELECT_TARGET,
        strict_two_fa_code=cnf.STRICT_TWO_FA_CODE,
        max_immediate_attempts=cnf.MAX_IMMEDIATE_ATTEMPTS,
        oauth_timeout=cnf.OAUTH_TIMEOUT,
        max_presubmit_buffer=cnf.MAX_PRESUBMIT_BUFFER,
        min_presubmit_buffer=cnf.MIN_PRESUBMIT_BUFFER,
        max_failed_auth=cnf.MAX_FAILED_AUTH,
        outputs_dir=cnf.OUTPUTS_DIR,
        use_paper_account=cnf.USE_PAPER_ACCOUNT,
    )

    process_handler = ProcessHandler(
        gateway_process_match=cnf.GATEWAY_PROCESS_MATCH,
        gateway_dir=cnf.GATEWAY_DIR,
        gateway_startup=cnf.GATEWAY_STARTUP,
        verify_connection=http_handler.base_route,
    )

    strategy_handler = StrategyHandler(
        http_handler=http_handler,
        login_handler=login_handler,
        process_handler=process_handler,
        authentication_strategy=cnf.AUTHENTICATION_STRATEGY,
        reauthenticate_wait=cnf.REAUTHENTICATE_WAIT,
        restart_failed_sessions=cnf.RESTART_FAILED_SESSIONS,
        restart_wait=cnf.RESTART_WAIT,
        max_reauthenticate_retries=cnf.MAX_REAUTHENTICATE_RETRIES,
        max_status_check_retries=cnf.MAX_STATUS_CHECK_RETRIES,
    )

    client = GatewayClient(
        http_handler=http_handler,
        strategy_handler=strategy_handler,
        process_handler=process_handler,
        health_server_port=cnf.HEALTH_SERVER_PORT,
        spawn_new_processes=cnf.SPAWN_NEW_PROCESSES,
        maintenance_interval=cnf.MAINTENANCE_INTERVAL,
        request_retries=cnf.REQUEST_RETRIES,
        active=cnf.START_ACTIVE,
    )

    def stop(_, _1):
        client.shutdown()
        sys.exit(0)

    signal.signal(signal.SIGINT, stop)
    signal.signal(signal.SIGTERM, stop)

    _LOGGER.info(f'Configuration:\n{cnf.all_variables}')

    if args.start:
        pids = process_handler.start_gateway()
        if pids is not None:
            _LOGGER.info(f'Gateway running with pids: {pids}')
        else:
            _LOGGER.info(f'Gateway not running.')
        while True:
            pass
    elif args.authenticate:
        success, _ = strategy_handler.try_authenticating()
        _LOGGER.info(f'Gateway {"" if success else "not "}authenticated.')
    elif args.check:
        status = http_handler.get_status()
        if not status.session:
            _LOGGER.info(f'No active Gateway session.')
        else:
            _LOGGER.info(f'Gateway session {"" if status.authenticated else "not "}authenticated.')
    elif args.tickle:
        success = http_handler.tickle().running
        _LOGGER.info(f'Gateway {"" if success else "not "}running.')
    elif args.maintain:
        client.maintain()
    elif args.kill:
        success = process_handler.kill_gateway()
        _LOGGER.info(f'Gateway {"" if success else "not "}killed.')
    else:
        # we have to do this here first because APS waits before running it the first time
        if client.active:
            success, shutdown, status = client.start_and_authenticate()
            if success:
                _LOGGER.info(f'Gateway running and authenticated, session id: {status.session_id}, server name: {status.server_name}')

            if shutdown:
                _LOGGER.warning('Shutting IBeam down due to critical error.')
            else:
                client.maintain()
        else:
            _LOGGER.info(f'IBeam initialised in an inactive state. Starting maintenance loop.')
            client.maintain()



import logging
import sys
import time

from pathlib import Path
from apscheduler.executors.pool import ThreadPoolExecutor, ProcessPoolExecutor
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.interval import IntervalTrigger

from ibeam.src.health_server import new_health_server
from ibeam.src.handlers.http_handler import HttpHandler, Status
from ibeam.src.handlers.process_handler import ProcessHandler
from ibeam.src.handlers.strategy_handler import StrategyHandler

sys.path.insert(0, str(Path(__file__).parent.parent))

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)

class GatewayClient():

    def __init__(self,
                 http_handler: HttpHandler,
                 strategy_handler: StrategyHandler,
                 process_handler: ProcessHandler,
                 health_server_port: int,
                 spawn_new_processes: bool,
                 maintenance_interval: int,
                 request_retries: int,
                 active:bool=True,
                 ):

        self._should_shutdown = False

        self.http_handler = http_handler
        self.strategy_handler = strategy_handler
        self.process_handler = process_handler

        self.health_server_port = health_server_port
        self.spawn_new_processes = spawn_new_processes
        self.maintenance_interval = maintenance_interval
        self.request_retries = request_retries

        self._concurrent_maintenance_attempts = 1
        self._health_server = new_health_server(
            self.health_server_port,
            self.http_handler.get_status,
            self.get_shutdown_status,
            self.on_activate,
            self.on_deactivate,
        )

        self._active = active

    def get_shutdown_status(self) -> bool:
        return self._should_shutdown

    def start_and_authenticate(self, request_retries=1) -> (bool, bool, Status):
        """Starts the gateway and authenticates using the credentials stored."""

        self.process_handler.start_gateway()

        success, shutdown, status = self.strategy_handler.try_authenticating(request_retries=request_retries)
        self._should_shutdown = shutdown
        return success, shutdown, status

    def on_activate(self) -> bool:
        if self._active:
            return True

        _LOGGER.info('Activating')
        self._active = True
        return True

    def on_deactivate(self) -> bool:
        if not self._active:
            return True

        _LOGGER.info('Deactivating')
        self._active = False
        self.http_handler.logout()
        self.process_handler.kill_gateway()
        return True

    def build_scheduler(self):
        if self.spawn_new_processes:
            executors = {'default': ProcessPoolExecutor(self._concurrent_maintenance_attempts)}
        else:
            executors = {'default': ThreadPoolExecutor(self._concurrent_maintenance_attempts)}
        job_defaults = {'coalesce': False, 'max_instances': self._concurrent_maintenance_attempts}
        self._scheduler = BackgroundScheduler(executors=executors, job_defaults=job_defaults, timezone='UTC')
        self._scheduler.add_job(self._maintenance, trigger=IntervalTrigger(seconds=self.maintenance_interval))

    def maintain(self):
        self.build_scheduler()
        _LOGGER.info(f'Starting maintenance with interval {self.maintenance_interval} seconds')
        self._scheduler.start()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt as e:
            _LOGGER.info('Keyboard interrupt, shutting down.')
            pass
        self._scheduler.remove_all_jobs()
        self._scheduler.shutdown(wait=False)
        if self._health_server:
            self._health_server.shutdown()

    def _maintenance(self):
        if not self._active:
            _LOGGER.info('Maintenance skipped, GatewayClient is not active.')
            return

        _LOGGER.info('Maintenance')

        success, shutdown, status = self.start_and_authenticate(request_retries=self.request_retries)

        if shutdown:
            _LOGGER.warning('Shutting IBeam down due to critical error.')
            self._scheduler.remove_all_jobs()
            self._scheduler.shutdown(False)
            if self._health_server:
                self._health_server.shutdown()
        elif success:
            _LOGGER.info(f'Gateway running and authenticated, session id: {status.session_id}, server name: {status.server_name}')
            validate_success = self.http_handler.validate()
            if not validate_success:
                _LOGGER.warning(f'Validation result is False when IBeam attempted to extend the SSO token. This could indicate token authentication issues.')

    def shutdown(self):
        self._health_server.shutdown()

    def __getstate__(self):
        state = self.__dict__.copy()

        # APS schedulers and health_server can't be pickled
        del state['_scheduler']
        del state['_health_server']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.build_scheduler()

    @property
    def active(self):
        return self._active



import logging
import os
import shutil
from pathlib import Path

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)


class InputsHandler():

    def __init__(self,
                 inputs_dir: str,
                 gateway_dir: str):

        self._cacert_jks_path = os.path.join(inputs_dir, 'cacert.jks')
        self._cacert_pem_path = os.path.join(inputs_dir, 'cacert.pem')

        self._valid_certificates = os.path.isfile(self._cacert_jks_path) and os.path.isfile(self._cacert_pem_path)
        if self._valid_certificates:
            _LOGGER.info('TLS certificates found and will be used for verification')

        gateway_root_dir = os.path.join(gateway_dir, 'root')

        config_source = os.path.join(inputs_dir, 'conf.yaml')
        if os.path.isfile(config_source):
            _LOGGER.info('Custom conf.yaml found and will be used by the Gateway')
            config_target = os.path.join(gateway_root_dir, 'conf.yaml')
            shutil.copy2(config_source, config_target)

        if self._valid_certificates:
            cacert_target = os.path.join(gateway_root_dir, os.path.basename(self._cacert_jks_path))
            shutil.copy2(self._cacert_jks_path, cacert_target)

    @property
    def cacert_jks_path(self):
        return self._cacert_jks_path

    @property
    def cacert_pem_path(self):
        return self._cacert_pem_path

    @property
    def valid_certificates(self):
        return self._valid_certificates


import logging
import os
import sys
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Optional

from pyvirtualdisplay import Display
from selenium import webdriver
from selenium.common import WebDriverException
from selenium.webdriver.chrome.service import Service

import ibeam
from ibeam.src import var

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)

_DRIVER_NAMES = {}


def _new_chrome_driver(driver_path, name: str = 'default', headless: bool = True, incognito: bool = True, ui_scaling: float = 1) -> webdriver.Chrome:
    """Creates a new chrome driver."""

    global _DRIVER_NAMES

    _DRIVER_NAMES[name] = True  # just to ensure the name is in the dict
    driver_index = list(_DRIVER_NAMES.keys()).index(name)  # order of insertion dictates the driver_index

    options = webdriver.ChromeOptions()
    if headless:
        options.add_argument('--headless')
        options.add_argument('--disable-gpu')
    options.add_argument('--no-sandbox')
    if incognito:
        options.add_argument("--incognito")  # this allows 2FA method to be selected every time
    options.add_argument('--ignore-ssl-errors=yes')
    options.add_argument('--ignore-certificate-errors')
    options.add_argument(f'--remote-debugging-port={9222 + driver_index}')
    options.add_argument('--useAutomationExtension=false')
    options.add_argument('--disable-extensions')
    options.add_argument('--dns-prefetch-disable')
    options.add_argument('--disable-features=VizDisplayCompositor')
    options.add_argument(f"--force-device-scale-factor={ui_scaling}")
    options.add_argument(f'--user-data-dir={tempfile.gettempdir()}/ibeam-chrome-{name}')
    service = Service(executable_path=driver_path)
    driver = webdriver.Chrome(options=options, service=service)
    if driver is None:
        _LOGGER.error('Unable to create a new chrome driver.')

    return driver


def release_chrome_driver(driver: webdriver.Chrome):
    driver.quit()


def start_driver(driver_path: str,
                 name: str = 'default',
                 headless: bool = True,
                 incognito: bool = True,
                 ui_scaling: float = 1,
                 page_load_timeout: int = 15
                 ) -> Optional[webdriver.Chrome]:
    try:
        driver = _new_chrome_driver(driver_path=driver_path, name=name, headless=headless, incognito=incognito, ui_scaling=ui_scaling)
        driver.set_page_load_timeout(page_load_timeout)
    except WebDriverException as e:
        if 'net::ERR_CONNECTION_REFUSED' in e.msg:
            _LOGGER.error(
                'Connection to Gateway refused. This could indicate IB Gateway is not running. Consider increasing IBEAM_GATEWAY_STARTUP wait buffer')
            return None
        if 'net::ERR_CONNECTION_CLOSED' in e.msg:
            _LOGGER.error(
                f'Connection to Gateway failed. This could indicate IB Gateway is not running correctly or that its port was already occupied')
            return None
        else:
            raise e

    return driver


def save_screenshot(driver, outputs_dir, postfix=''):
    if not var.ERROR_SCREENSHOTS or driver is None:
        return

    now = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    outputs_path = Path(outputs_dir)
    screenshot_name = f'ibeam__{ibeam.__version__}__{now}{postfix}.png'

    try:
        required_width = driver.execute_script('return document.body.parentNode.scrollWidth')
        required_height = driver.execute_script('return document.body.parentNode.scrollHeight')
        driver.set_window_size(required_width, required_height)

        outputs_path.mkdir(exist_ok=True)
        screenshot_filepath = os.path.join(outputs_dir, screenshot_name)

        # a little hack to prevent overwriting screenshots saved in the same second
        if os.path.exists(screenshot_filepath):
            save_screenshot(driver, outputs_dir, postfix + '_')
            return

        _LOGGER.info(
            f'Saving screenshot to {screenshot_filepath}. Make sure to cover your credentials if you share it with others.')
        driver.get_screenshot_as_file(screenshot_filepath)
    except Exception as e:
        _LOGGER.exception(f"Exception while saving screenshot: {str(e)} for screenshot: {screenshot_name}")


class DriverFactory():
    def __init__(self,
                 driver_path: str,
                 name: str = 'default',
                 headless: bool = True,
                 incognito: bool = True,
                 ui_scaling: float = 1,
                 page_load_timeout: int = 15
                 ):
        self.driver_path = driver_path
        self.name = name
        self.headless = headless
        self.incognito = incognito
        self.ui_scaling = ui_scaling
        self.page_load_timeout = page_load_timeout

    def new_driver(self,
                   driver_path: str = None,
                   name: str = None,
                   headless: bool = None,
                   incognito: bool = None,
                   ui_scaling: float = None,
                   page_load_timeout: int = None
                   ) -> webdriver.Chrome:

        driver_path = driver_path if driver_path is not None else self.driver_path
        name = name if name is not None else self.name
        headless = headless if headless is not None else self.headless
        incognito = incognito if incognito is not None else self.incognito
        ui_scaling = ui_scaling if ui_scaling is not None else self.ui_scaling
        page_load_timeout = page_load_timeout if page_load_timeout is not None else self.page_load_timeout

        return start_driver(driver_path=driver_path, name=name, headless=headless, incognito=incognito, ui_scaling=ui_scaling, page_load_timeout=page_load_timeout)


def start_up_browser(driver_factory:DriverFactory) -> (webdriver.Chrome, Optional[Display]):
    display = None
    if sys.platform == 'linux':
        display = Display(visible=False, size=(800, 600))
        display.start()

    driver = driver_factory.new_driver()
    if driver is None:
        return False, False

    return driver, display


def shut_down_browser(driver:webdriver.Chrome, display:Optional[Display]):
    _LOGGER.info(f'Cleaning up the resources. Display: {display} | Driver: {driver}')

    if display is not None:
        display.stop()

    if driver is not None:
        release_chrome_driver(driver)



from abc import ABC, abstractmethod
from typing import Optional

from selenium import webdriver

class TwoFaHandler(ABC):

    def __init__(self, outputs_dir:str):
        self.outputs_dir = outputs_dir

    @abstractmethod
    def get_two_fa_code(self, driver:webdriver.Chrome) -> Optional[str]:
        raise NotImplementedError()

    def __str__(self):
        return "TwoFaHandler()"



import threading
from http.server import HTTPServer, BaseHTTPRequestHandler
from pathlib import Path
import logging

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)


def new_health_server(port: int, check_status, get_shutdown_status,
                      activate_callback:callable,
                      deactivate_callback:callable):
    class HealthzHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            if self.path == "/livez":
                return self._live()
            elif self.path == "/readyz":
                return self._ready()
            elif self.path == "/activate":
                return self._activate()
            elif self.path == "/deactivate":
                return self._deactivate()
            self.send_error(404, "Not Found")

        def _live(self):
            if get_shutdown_status():
                self._send_500()
            else:
                self._send_ok()

        def _ready(self):
            status = check_status()
            if not status.authenticated:
                return self._not_ready()
            self._send_ok()

        def _activate(self):
            if activate_callback():
                return self._send_ok()
            else:
                return self._send_500()

        def _deactivate(self):
            if deactivate_callback():
                return self._send_ok()
            else:
                return self._send_500()

        def _send_ok(self):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write("OK".encode())

        def _send_500(self):
            self.send_response(500)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write("Internal Error".encode())

        def _not_ready(self):
            self.send_response(503)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write("Not Ready".encode())

    server = HTTPServer(('', port), HealthzHandler)
    threading.Thread(target=server.serve_forever).start()
    _LOGGER.info(f'Health server started at port={port}')
    return server



import logging
from pathlib import Path
from typing import Optional

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC

from ibeam.config import Config
from ibeam.src.utils.selenium_utils import text_to_be_present_in_element

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)


class Target():
    def __init__(
            self,
            variable: str,
            ):
        type, identifier = variable.split("@@")
        self.type = type
        self.identifier = identifier
        self.variable = variable

        if type == 'ID':
            self.by = By.ID
            self._identify = self.identify_by_id
        elif type == 'CSS_SELECTOR':
            self.by = By.CSS_SELECTOR
            self._identify = self.identify_by_css_selector
        elif type == 'CLASS_NAME':
            self.by = By.CLASS_NAME
            self._identify = self.identify_by_class
        elif type == 'NAME':
            self.by = By.NAME
            self._identify = self.identify_by_name
        elif type == 'FOR':
            self.by = By.CSS_SELECTOR
            self._identify = self.identify_by_for
        elif type == 'TAG_NAME':
            self.by = [(By.TAG_NAME, 'pre'), (By.TAG_NAME, 'body')]
            self._identify = self.identify_by_text
        else:
            raise RuntimeError(f'Unknown target type: {type}@@{identifier}')

    def identify(self, trigger: WebElement) -> bool:
        return self._identify(trigger)

    def identify_by_id(self, trigger: WebElement) -> bool:
        return self.identifier in trigger.get_attribute('id')

    def identify_by_css_selector(self, trigger: WebElement) -> bool:
        return self.identifier.replace('.', ' ').strip() in trigger.get_attribute('class')

    def identify_by_class(self, trigger: WebElement) -> bool:
        return self.identifier in trigger.get_attribute('class')

    def identify_by_name(self, trigger: WebElement) -> bool:
        return self.identifier in trigger.get_attribute('name')

    def identify_by_text(self, trigger: WebElement) -> bool:
        return self.identifier in trigger.text

    def identify_by_for(self, trigger: WebElement) -> bool:
        return self.identifier in trigger.get_attribute('for')

    def __repr__(self):
        return f'Target({self.variable})'


Targets = dict[str, Target]


def targets_from_versions(targets: Targets, versions: dict) -> Targets:
    version_target_user_name = Target(versions['USER_NAME_EL'])
    version_target_error = Target(versions['ERROR_EL'])

    if 'USER_NAME' in targets and version_target_user_name.variable != targets['USER_NAME'].variable:
        _LOGGER.warning(f'USER_NAME target is forced to "{targets["USER_NAME"].variable}", contrary to the element found on the website: "{version_target_user_name}"')
    else:
        targets['USER_NAME'] = version_target_user_name

    if "ERROR" in targets and version_target_error.variable != targets['ERROR'].variable:
        _LOGGER.warning(f'ERROR target is forced to "{targets["ERROR"].variable}", contrary to the element found on the website: "{version_target_error}"')
    else:
        targets['ERROR'] = version_target_error

    return targets


def create_targets(cnf: Config) -> Targets:
    targets = {}

    targets['PASSWORD'] = Target(cnf.PASSWORD_EL)
    targets['SUBMIT'] = Target(cnf.SUBMIT_EL)
    targets['SUCCESS'] = Target(cnf.SUCCESS_EL_TEXT)
    targets['IBKEY_PROMO'] = Target(cnf.IBKEY_PROMO_EL_CLASS)
    targets['TWO_FA'] = Target(cnf.TWO_FA_EL_ID)
    targets['TWO_FA_NOTIFICATION'] = Target(cnf.TWO_FA_NOTIFICATION_EL)
    targets['TWO_FA_INPUT'] = Target(cnf.TWO_FA_INPUT_EL_ID)
    targets['TWO_FA_SELECT'] = Target(cnf.TWO_FA_SELECT_EL_ID)
    targets['LIVE_PAPER_TOGGLE'] = Target(cnf.LIVE_PAPER_TOGGLE_EL)

    return targets


def identify_target(trigger: WebElement, targets: Targets) -> Optional[Target]:
    for target in targets.values():
        try:
            if target.identify(trigger):
                return target
        except TypeError as e:
            # this is raised if trigger doesn't have Target's attribute, we can ignore it
            if "argument of type 'NoneType' is not iterable" in str(e):
                continue
            raise

    raise RuntimeError(f'Trigger found but cannot be identified: {trigger} :: {trigger.get_attribute("outerHTML")}')


def is_present(target: Target) -> callable:
    return EC.presence_of_element_located((target.by, target.identifier))


def is_visible(target: Target) -> callable:
    return EC.visibility_of_element_located((target.by, target.identifier))


def is_clickable(target: Target) -> callable:
    return EC.element_to_be_clickable((target.by, target.identifier))


def has_text(target: Target) -> callable:
    return text_to_be_present_in_element(target.by, target.identifier)


def find_element(target: Target, driver: webdriver.Chrome) -> WebElement:
    return driver.find_element(target.by, target.identifier)



import importlib
import importlib.util
import os
from pathlib import Path
from typing import Optional

from ibeam.src.login.driver import DriverFactory
from ibeam.src.two_fa_handlers.external_request_handler import ExternalRequestTwoFaHandler
from ibeam.src.two_fa_handlers.google_msg_handler import GoogleMessagesTwoFaHandler
from ibeam.src.two_fa_handlers.notification_resend_handler import NotificationResendTwoFaHandler
from ibeam.src.two_fa_handlers.two_fa_handler import TwoFaHandler


def select(handler_name:str,
           driver_factory:DriverFactory,
           outputs_dir:str,
           custom_two_fa_handler:str,
           inputs_dir: os.PathLike) -> Optional[TwoFaHandler]:

    if handler_name == 'GOOGLE_MSG':
        handler = GoogleMessagesTwoFaHandler(driver_factory, outputs_dir=outputs_dir)
    elif handler_name == 'EXTERNAL_REQUEST':
        handler = ExternalRequestTwoFaHandler(outputs_dir=outputs_dir)
    elif handler_name == 'NOTIFICATION_RESEND':
        handler = NotificationResendTwoFaHandler(outputs_dir=outputs_dir)
    elif handler_name == 'CUSTOM_HANDLER':
        handler = load_custom_two_fa_handler(custom_two_fa_handler, inputs_dir)(outputs_dir=outputs_dir)
    else:
        handler = None

    return handler


def load_custom_two_fa_handler(two_fa_handler_fqp, inputs_dir: os.PathLike):
    module_name, class_name = two_fa_handler_fqp.rsplit('.', 1)
    handler_filepath = Path(inputs_dir, module_name + '.py')
    try:
        spec = importlib.util.spec_from_file_location(module_name, os.fspath(handler_filepath))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except ModuleNotFoundError as e:
        if str(e) == f'No module named \'{module_name}\'':
            raise RuntimeError(
                f'Invalid handler path: "{two_fa_handler_fqp}". Module "{module_name}" not found. You need to provide the MODULE_NAME.CLASS_NAME of your custom handler as IBEAM_CUSTOM_TWO_FA_HANDLER environment variable.') from e
        else:
            raise e

    try:
        klass = getattr(module, class_name)
    except AttributeError as e:
        if str(e) == f'module \'{module_name}\' has no attribute \'{class_name}\'':
            raise RuntimeError(
                f'Invalid handler path: "{two_fa_handler_fqp}". Module "{module_name}" has no class "{class_name}". You need to provide the MODULE_NAME.CLASS_NAME of your custom handler as IBEAM_CUSTOM_TWO_FA_HANDLER environment variable.') from e
        else:
            raise e

    return klass



import logging
import os
import subprocess
import sys
import time
from pathlib import Path
from typing import Optional, List

import psutil

from ibeam.src.handlers.http_handler import HttpHandler

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)


# from https://stackoverflow.com/questions/550653/cross-platform-way-to-get-pids-by-process-name-in-python/2241047
def _find_procs_by_name(name):
    "Return a list of processes matching 'name'."
    assert name, name
    ls = []
    for p in psutil.process_iter():
        name_, exe, cmdline = "", "", []
        try:
            # name_ = p.name()
            cmdline = p.cmdline()
            exe = p.exe()
        except (psutil.AccessDenied, psutil.ZombieProcess):
            pass
        except psutil.NoSuchProcess:
            continue
        if name in ' '.join(cmdline) or name in os.path.basename(exe):
            ls.append(p)
    return ls


def _start_gateway(gateway_dir):
    creationflags = 0  # when not on Windows, we send 0 to avoid errors.

    if sys.platform == 'win32':
        args = ["cmd", "/k", r"bin\run.bat", r"root\conf.yaml"]
        _LOGGER.info(f'Starting Gateway as Windows process with params: {args}')
        creationflags = subprocess.CREATE_NEW_CONSOLE

    elif sys.platform == 'darwin':
        args = ["open", "-F", "-a", "Terminal", r"bin/run.sh", r"root/conf.yaml"]
        _LOGGER.info(f'Starting Gateway as Mac process with params: {args}')

    elif sys.platform == 'linux':
        args = ["bash", r"bin/run.sh", r"root/conf.yaml"]
        _LOGGER.info(f'Starting Gateway as Linux process with params: {args}')

    else:
        raise EnvironmentError(f'Unknown platform: {sys.platform}')

    subprocess.Popen(
        args=args,
        cwd=gateway_dir,
        creationflags=creationflags
    )

def _try_starting_gateway(
        gateway_process_match:str,
        gateway_dir:os.PathLike,
        gateway_startup:int,
        verify_connection:callable,
)  -> Optional[List[int]]:
    processes = _find_procs_by_name(gateway_process_match)
    if processes:
        server_process_pids = [process.pid for process in processes]
    else:
        _LOGGER.info('Gateway not found, starting new one...')
        _LOGGER.info(
            'Note that the Gateway log below may display "Open https://localhost:[PORT] to login" - ignore this command.')

        _start_gateway(gateway_dir)

        server_process_pids = None

        # let's try to communicate with the Gateway
        t_end = time.time() + gateway_startup

        while time.time() < t_end:
            processes = _find_procs_by_name(gateway_process_match)
            if len(processes) == 0:
                continue

            server_process_pids = [process.pid for process in processes]
            _LOGGER.info(f'Gateway started with pids: {server_process_pids}')
            break

        if server_process_pids is None:
            _LOGGER.error(f'Cannot find gateway process by name: "{gateway_process_match}"')
            return None

        ping_success = False
        while time.time() < t_end:
            status = verify_connection()
            if not status.running:
                seconds_remaining = round(t_end - time.time())
                if seconds_remaining > 0:
                    _LOGGER.info(
                        f'Cannot ping Gateway. Retrying for another {seconds_remaining} seconds')
                    time.sleep(1)
            else:
                _LOGGER.info('Gateway connection established')
                ping_success = True
                break

        if not ping_success:
            _LOGGER.error('Gateway process found but cannot establish a connection with the Gateway')

    return server_process_pids

def _kill_gateway(gateway_process_match:str):
    processes = _find_procs_by_name(gateway_process_match)
    if not processes:
        _LOGGER.warning(f'Attempting to kill but could not find process named "{gateway_process_match}"')
        return False

    for process in processes:
        process.terminate()

    time.sleep(1)

    # double check we succeeded
    processes = _find_procs_by_name(gateway_process_match)
    if processes:
        return False
    return True


class ProcessHandler():

    def __init__(self,
                 gateway_dir: os.PathLike,
                 gateway_process_match:str,
                 gateway_startup:int,
                 verify_connection:callable,
                 ):
        self.gateway_dir = gateway_dir
        self.gateway_process_match = gateway_process_match
        self.gateway_startup = gateway_startup
        self.verify_connection = verify_connection

    def start_gateway(self) ->  Optional[List[int]]:
        return _try_starting_gateway(
            gateway_process_match=self.gateway_process_match,
            gateway_dir=self.gateway_dir,
            gateway_startup=self.gateway_startup,
            verify_connection=self.verify_connection,
        )

    def kill_gateway(self) -> bool:
        return _kill_gateway(gateway_process_match=self.gateway_process_match)


import logging
import os
import time
from pathlib import Path
from typing import Optional

from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait

from ibeam.src.utils.selenium_utils import text_to_be_present_in_element
from ibeam.src.login.driver import save_screenshot
from ibeam.src.two_fa_handlers.two_fa_handler import TwoFaHandler
from selenium.webdriver.support import expected_conditions as EC

_NOTIFICATION_RESEND_RETRIES = int(os.environ.get('IBEAM_NOTIFICATION_RESEND_RETRIES', 10))
"""How many times to resend the notification."""

_NOTIFICATION_RESEND_INTERVAL = int(os.environ.get('IBEAM_NOTIFICATION_RESEND_INTERVAL', 10))
"""How many times to resend the notification."""

_NOTIFICATION_RESEND_EL = os.environ.get('IBEAM_NOTIFICATION_RESEND_EL', "a[onclick*='resendNotification()']")
"""Css selector for the resend notification button."""

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)


class NotificationResendTwoFaHandler(TwoFaHandler):
    """ This 2FA handler will repeatedly resend notifications to user's phone. """

    def check_and_resend(self, driver:webdriver.Chrome, success_text:str, depth=0):
        if depth >= _NOTIFICATION_RESEND_RETRIES:
            _LOGGER.error(f'Reached maximum number of notification resend retries: {_NOTIFICATION_RESEND_RETRIES}. Aborting.')
            return False

        try:
            WebDriverWait(driver, 30).until(EC.element_to_be_clickable((By.CSS_SELECTOR, _NOTIFICATION_RESEND_EL)))
        except TimeoutException:
            _LOGGER.error(f'Notification resend element not found: {_NOTIFICATION_RESEND_EL}. Aborting.')
            return False

        notification_resend_el = driver.find_element_by_css_selector(_NOTIFICATION_RESEND_EL)
        notification_resend_el.click()

        success_present = text_to_be_present_in_element([(By.TAG_NAME, 'pre'), (By.TAG_NAME, 'body')], success_text)

        try:
            WebDriverWait(driver, _NOTIFICATION_RESEND_INTERVAL).until(success_present)
        except TimeoutException:
            _LOGGER.info(f'Success condition was not found when resending 2FA notification. Reattempting {_NOTIFICATION_RESEND_RETRIES - depth - 1} more times.')
            return self.check_and_resend(driver, success_text, depth + 1)
        else:
            return True

    def get_two_fa_code(self, driver:webdriver.Chrome) -> Optional[bool]:
        raise NotImplementedError()

    def interact_with_notification(self, driver:webdriver.Chrome, success_text:str) -> Optional[bool]:
        time.sleep(2)  # give the first notification a short while to arrive first
        try:
            return self.check_and_resend(driver, success_text)
        except Exception as e:
            _LOGGER.exception(f'Exception while handling notification resend 2FA: {e}')
            save_screenshot(driver, self.outputs_dir, postfix='__notification_2fa')

    def __str__(self):
        return "NotificationResendTwoFaHandler()"



from typing import Union

from selenium.common import StaleElementReferenceException, WebDriverException
from selenium.webdriver.remote.webelement import WebElement


class text_to_be_present_in_element(object):
    """ An expectation for checking if the given text is present in the
    specified element.
    locator, text
    """

    def __init__(self, locators, text_):
        if not isinstance(locators, list):
            locators = locators
        self.locators = locators
        self.text = text_

    def __call__(self, driver):
        for locator in self.locators:
            try:
                element = driver.find_element(*locator)
                if self.text in element.text:
                    return element
            except StaleElementReferenceException:
                continue
        return False


def any_of(*expected_conditions):
    """ An expectation that any of multiple expected conditions is true.
    Equivalent to a logical 'OR'.
    Returns results of the first matching condition, or False if none do. """

    def any_of_condition(driver) -> Union[WebElement, bool]:
        for expected_condition in expected_conditions:
            try:
                result = expected_condition(driver)
                if result:
                    return result
            except WebDriverException:
                pass
        return False

    return any_of_condition



from distutils.core import setup

from setuptools import find_packages

import ibeam

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    parsed_requirements = f.read().splitlines()

pakages_found = find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests", "examples", "docs", "out", "dist"])

print(f'Found packages: {pakages_found}')

setup(
    name='ibeam',
    packages=pakages_found,
    version=ibeam.__version__,
    license='Apache-2.0',
    description='IBeam is an authentication and maintenance tool used for the Interactive Brokers Client Portal Web API Gateway.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Voy Zan',
    author_email='voy1982@yahoo.co.uk',
    url='https://github.com/Voyz/ibeam',
    keywords=['interactive brokers', 'algo trading',
              'algorithmic trading', 'data flow', 'quant', 'trading'],
    install_requires=parsed_requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.7',
    ],
)



import json
import logging
import socket
import ssl
from pathlib import Path
from urllib import request

from urllib.error import HTTPError, URLError
import urllib.request
import urllib.parse

from ibeam.src.handlers.inputs_handler import InputsHandler
from ibeam.src.utils.py_utils import exception_to_string

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)


class Status():
    """
    A class to represent the status of the Gateway and our connection to IBKR servers.

    Attributes:
        running (bool): Whether the Gateway is running.
        session (bool): Whether there is an active session.
        response (dict): The server response to the request.
        connected (bool): Whether the server reports 'connected' to the Gateway.
        authenticated (bool): Whether the session is authenticated.
        competing (bool): Whether there are competing sessions on the server.
        collision (bool): Currently no idea what this flag means.
        session_id (str): The ID of the current session, if one exists.
        server_name (str): The name of the server.
        server_version (str): The version of the server.
        expires (int): The expiration time of the current session, represented as milliseconds.

    """
    def __init__(self,
                 running: bool = False,
                 session: bool = False,
                 response = None,

                 # parsed args
                 connected: bool = False,
                 authenticated: bool = False,
                 competing: bool = False,
                 collision: bool = False,
                 session_id: str = None,
                 server_name: str = None,
                 server_version: str = None,
                 expires: int = None,
                 ):
        self.running = running
        self.session = session
        self.response = response

        # parsed args
        self.connected = connected
        self.authenticated = authenticated
        self.competing = competing
        self.collision = collision
        self.session_id = session_id
        self.server_name = server_name
        self.server_version = server_version
        self.expires = expires

    def expiration_time(self):
        if self.expires is None:
            return None

        return f'{int(self.expires / 1000)} seconds'

    @property
    def parsed_status(self):
        if not self.running:
            return 'NOT RUNNING'
        if not self.session:
            return 'NO SESSION'
        if not self.connected:
            return 'NOT CONNECTED'
        if self.competing:
            return 'COMPETING'
        if self.collision:
            return 'COLLISION'
        if self.authenticated:
            return 'AUTHENTICATED'

        return 'NOT AUTHENTICATED'

    def __repr__(self):
        d = self.__dict__.copy()
        if 'response' in d:
            d.pop('response')
        return f'Status({", ".join([f"{k}={repr(v)}" for k, v in d.items()])})'

    def __str__(self):
        return f'{self.parsed_status} {repr(self)}'
class HttpHandler():

    def __init__(self,
                 inputs_handler: InputsHandler,
                 base_url:str,
                 route_validate: str,
                 route_tickle: str,
                 route_logout: str,
                 route_reauthenticate: str,
                 route_initialise: str,
                 request_timeout: int,
                 ):

        self.inputs_handler = inputs_handler
        self.base_url = base_url

        self.route_validate = route_validate
        self.route_tickle = route_tickle
        self.route_logout = route_logout
        self.route_reauthenticate = route_reauthenticate
        self.route_initialise = route_initialise
        self.request_timeout = request_timeout
        self.build_ssh_context()

    def build_ssh_context(self):
        self.ssl_context = ssl.SSLContext()
        if self.inputs_handler.valid_certificates:
            self.ssl_context.verify_mode = ssl.CERT_REQUIRED
            self.ssl_context.check_hostname = True
            self.ssl_context.load_verify_locations(self.inputs_handler.cacert_pem_path)

    def url_request(self, url, method='GET'):
        _LOGGER.debug(f'{method} {url}{"" if self.inputs_handler.valid_certificates else " (unverified)"}')
        req = request.Request(url, method=method)
        return urllib.request.urlopen(req, context=self.ssl_context, timeout=self.request_timeout)

    def try_request(self, url, method='GET', max_attempts=1) -> Status:
        """Attempts a HTTP request and returns Status object indicating whether the gateway can be reached, whether there is an active session and whether it is authenticated. Attempts to repeat the request up to max_attempts times.

        status.running -> gateway running
        status.session -> active session present
        status.authenticated -> session authenticated (equivalent to 'all good')
        status.competing -> session competing
        """

        def _request(attempt=0) -> Status:
            status = Status()
            try:
                # if this doesn't throw an exception, the gateway is running and there is an active session
                response = self.url_request(url, method=method)
                status.running = True
                status.session = True
                status.response = response

                return status

            except HTTPError as e:
                status.running = True

                if e.code == 401:
                    # the gateway is running but there is no active session
                    pass

                elif e.code == 500 and 'Internal Server Error' in str(e):
                    _LOGGER.error(f'IBKR server error: "{e}". One of reasons for this error is IBKR server restart.')
                    status.session = False  # ensure we reauthenticate


                elif e.code == 503 and 'Service Unavailable' in str(e):
                    _LOGGER.error(f'IBKR service unavailable: "{e}". It seems IBKR servers are not ready to handle requests. We may need to wait until the servers are ready.')
                    status.session = False  # ensure we reauthenticate

                else:  # todo: possibly other codes could appear when not authenticated, fix when necessary
                    try:
                        raise RuntimeError('Unrecognised HTTPError') from e
                    except Exception as ee:
                        _LOGGER.exception(ee)

            except (URLError, socket.timeout) as e:
                reason = str(e)

                """
                    No connection... - happens when port isn't open
                    Cannot assign... - happens when calling a port taken by Docker but not served, when called from within the container.
                    Errno 0... - happens when calling a port taken by Docker but not served, when called from the host machine.
                """

                if 'No connection could be made because the target machine actively refused it' in reason \
                        or 'Cannot assign requested address' in reason \
                        or '[Errno 0] Error' in reason:
                    pass  # we expect these errors and don't need to log them

                elif "timed out" in reason \
                        or "The read operation timed out" in reason:
                    #TODO: this will cause full relogin, we probably only need to repeat the request
                    _LOGGER.error(
                        f'Connection timeout after {self.request_timeout} seconds. Consider increasing timeout by setting IBEAM_REQUEST_TIMEOUT environment variable. Error: {reason}')
                    status.running = True

                elif 'Connection refused' in reason:
                    _LOGGER.info(
                        f'Gateway running but not serving yet. Consider increasing IBEAM_GATEWAY_STARTUP timeout. Error: {reason}')
                    status.running = True

                elif 'An existing connection was forcibly closed by the remote host' in reason:
                    _LOGGER.error(
                        'Connection to Gateway was forcibly closed by the remote host. This means something is closing the Gateway process.')

                elif 'certificate verify failed: self signed certificate' in reason:
                    _LOGGER.error(
                        'Failed to verify the self-signed certificate. This could mean your self-generated .jks certificate and password are not correctly provided in Inputs Directory or listed in conf.yaml. Ensure your certificate\'s filename and password are correctly listed in conf.yaml, or see https://github.com/Voyz/ibeam/wiki/TLS-Certificates-and-HTTPS#certificates-in-confyaml for more information.')

                else:
                    try:
                        raise RuntimeError('Unrecognised URLError or socket.timeout') from e
                    except Exception as ee:
                        _LOGGER.exception(ee)
                    status.running = True

            except ConnectionResetError as e:
                if 'An existing connection was forcibly closed by the remote host' in str(e):
                    _LOGGER.error(
                        'Connection to Gateway was forcibly closed by the remote host. This means something is closing the Gateway process.')
                else:
                    try:
                        raise RuntimeError('Unrecognised ConnectionResetError') from e
                    except Exception as ee:
                        _LOGGER.exception(ee)

            except Exception as e:  # all other exceptions
                _LOGGER.exception(f'Unrecognised Exception:\n{exception_to_string(e)}')

            if max_attempts <= 1:
                return status
            else:
                if attempt >= max_attempts - 1:
                    _LOGGER.info(
                        f'Max request retries reached after {max_attempts} attempts. Consider increasing the retries by setting IBEAM_REQUEST_RETRIES environment variable')
                    return status
                else:
                    _LOGGER.info(f'Attempt number {attempt + 2}')
                    return _request(attempt + 1)

        return _request(0)

    def get_status(self, max_attempts=1) -> Status:
        """We use tickle instead of iserver/auth/status because it is more versatile."""
        status = self.tickle(max_attempts=max_attempts)

        if status.session:
            status.response = json.loads(status.response.read().decode('utf8'))

            status.authenticated = status.response['iserver']['authStatus']['authenticated']
            status.competing = status.response['iserver']['authStatus']['competing']
            status.connected = status.response['iserver']['authStatus']['connected']
            status.collision = status.response['collission']
            status.session_id = status.response['session']
            status.expires = int(status.response['ssoExpires'])

            # some fields are not present if unauthenticated
            status.server_name = status.response['iserver']['authStatus'].get('serverInfo', {}).get('serverName')
            status.server_version = status.response['iserver']['authStatus'].get('serverInfo', {}).get('serverVersion')

        return status

    def validate(self) -> bool:
        """Validate provides information on the current session. Works also after logout."""
        status = self.try_request(self.base_url + self.route_validate, 'GET')
        if status.session:
            status.response = json.loads(status.response.read().decode('utf8'))
            return status.response['RESULT']
        return False

    def tickle(self, max_attempts=1) -> Status:
        return self.try_request(self.base_url + self.route_tickle, 'POST', max_attempts=max_attempts)

    def logout(self):
        """Logout will log the user out, but maintain the session, allowing us to reauthenticate directly."""
        return self.url_request(self.base_url + self.route_logout, 'POST')

    def reauthenticate(self):
        """Reauthenticate will work only if there is an existing session."""
        return self.url_request(self.base_url + self.route_reauthenticate, 'POST')

    def initialise(self):
        """Initialise the session."""
        return self.url_request(self.base_url + self.route_initialise, 'POST')

    def base_route(self):
        """Call base base_url"""
        return self.try_request(self.base_url)


    def __getstate__(self):
        state = self.__dict__.copy()
        # ssl_context can't be pickled
        del state['ssl_context']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.build_ssh_context()



import os

from ibeam.src.utils.py_utils import strtobool

UNDEFINED = object()


def to_bool(value):
    return bool(strtobool(str(value)))


INPUTS_DIR = os.environ.get('IBEAM_INPUTS_DIR', '/srv/inputs/')
"""Directory path of Inputs Directory."""

OUTPUTS_DIR = os.environ.get('IBEAM_OUTPUTS_DIR',
                             os.path.abspath(
                                 os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..', 'outputs')))
"""Directory path of Outputs Directory."""

GATEWAY_DIR = os.environ.get('IBEAM_GATEWAY_DIR', UNDEFINED)
"""Path to the root of the IBKR Gateway."""

CHROME_DRIVER_PATH = os.environ.get('IBEAM_CHROME_DRIVER_PATH', UNDEFINED)
"""Path to the Chrome Driver executable file."""

GATEWAY_STARTUP = int(os.environ.get('IBEAM_GATEWAY_STARTUP', 20))
"""How many seconds to wait for the Gateway to respond after its startup."""

GATEWAY_PROCESS_MATCH = os.environ.get('IBEAM_GATEWAY_PROCESS_MATCH', 'ibgroup.web.core.clientportal.gw.GatewayStart')
"""The gateway process' name to match against."""

MAINTENANCE_INTERVAL = int(os.environ.get('IBEAM_MAINTENANCE_INTERVAL', 60))
"""How many seconds between each maintenance."""

SPAWN_NEW_PROCESSES = to_bool(os.environ.get('IBEAM_SPAWN_NEW_PROCESSES', False))
"""Whether new processes should be spawned for each maintenance."""

LOG_LEVEL = os.environ.get('IBEAM_LOG_LEVEL', 'INFO')
"""Verbosity level of the logger used."""

LOG_TO_FILE = to_bool(os.environ.get('IBEAM_LOG_TO_FILE', True))
"""Whether logs should also be saved to a file."""

LOG_FORMAT = os.environ.get('IBEAM_LOG_FORMAT', '%(asctime)s|%(levelname)-.1s| %(message)s')
"""Log format that is used by IBeam. """

REQUEST_RETRIES = int(os.environ.get('IBEAM_REQUEST_RETRIES', 2))
"""How many times to reattempt a request to the gateway."""

REQUEST_TIMEOUT = int(os.environ.get('IBEAM_REQUEST_TIMEOUT', 15))
"""How many seconds to wait for a request to complete."""

RESTART_FAILED_SESSIONS = to_bool(os.environ.get('IBEAM_RESTART_FAILED_SESSIONS', True))
"""Whether Gateway should be restarted on failed sessions."""

RESTART_WAIT = int(os.environ.get('IBEAM_RESTART_WAIT', 15))
"""How many seconds to wait for a restart to complete."""

REAUTHENTICATE_WAIT = int(os.environ.get('IBEAM_REAUTHENTICATE_WAIT', 15))
"""How many seconds to wait for a reauthentication to complete."""

HEALTH_SERVER_PORT = int(os.environ.get("IBEAM_HEALTH_SERVER_PORT", 5001))
"""Port to start health server on."""

SECRETS_SOURCE = os.environ.get("IBEAM_SECRETS_SOURCE", 'env')
"""Source of secrets."""

GCP_SECRETS_URL = os.environ.get("IBEAM_GCP_SECRETS_URL", None)
"""Base URL for GCP secrets manager."""

START_ACTIVE = to_bool(os.environ.get('IBEAM_START_ACTIVE', True))
"""Whether IBeam should start activated or dormant."""

########### GATEWAY ROUTES ###########

GATEWAY_BASE_URL = os.environ.get('IBEAM_GATEWAY_BASE_URL', "https://localhost:5000")
"""Base URL of the gateway."""

ROUTE_AUTH = os.environ.get('IBEAM_ROUTE_AUTH', '/sso/Login?forwardTo=22&RL=1&ip2loc=on')
"""Gateway route with authentication page."""

ROUTE_VALIDATE = os.environ.get('IBEAM_ROUTE_VALIDATE', '/v1/portal/sso/validate')
"""Gateway route with validation call."""

ROUTE_REAUTHENTICATE = os.environ.get('IBEAM_ROUTE_REAUTHENTICATE', '/v1/portal/iserver/reauthenticate?force=true')
"""Gateway route with reauthentication call."""

ROUTE_INITIALISE = os.environ.get('IBEAM_ROUTE_INITIALISE', '/v1/api/iserver/auth/ssodh/init')
"""Gateway route with session initialisation call."""

ROUTE_AUTH_STATUS = os.environ.get('IBEAM_ROUTE_AUTH_STATUS', '/v1/api/iserver/auth/status')
"""Gateway route with authentication status call."""

ROUTE_TICKLE = os.environ.get('IBEAM_ROUTE_TICKLE', '/v1/api/tickle')
"""Gateway route with tickle call."""

ROUTE_LOGOUT = os.environ.get('IBEAM_ROUTE_LOGOUT', '/v1/api/logout')
"""Gateway route with logout call."""

########### AUTHENTICATION ###########

USER_NAME_EL = os.environ.get('IBEAM_USER_NAME_EL', None)
"""HTML element name attribute containing the username input field."""

PASSWORD_EL = os.environ.get('IBEAM_PASSWORD_EL', 'NAME@@password')
"""HTML element name attribute containing the password input field."""

SUBMIT_EL = os.environ.get('IBEAM_SUBMIT_EL', 'CSS_SELECTOR@@.btn.btn-lg.btn-primary')
"""HTML element name attribute containing the submit button."""

ERROR_EL = os.environ.get('IBEAM_ERROR_EL', None)
"""HTML element class name attribute containing the submit button."""

SUCCESS_EL_TEXT = os.environ.get('IBEAM_SUCCESS_EL_TEXT', 'TAG_NAME@@Client login succeeds')
"""HTML element text indicating successful authentication."""

LIVE_PAPER_TOGGLE_EL = os.environ.get('IBEAM_LIVE_PAPER_TOGGLE_EL', 'FOR@@label[for=toggle1]')
"""HTML element name attribute containing the live paper toggle button."""

USE_PAPER_ACCOUNT = to_bool(os.environ.get('IBEAM_USE_PAPER_ACCOUNT', False))
"""Whether to attempt switching to paper mode before attempting logging in."""

OAUTH_TIMEOUT = int(os.environ.get('IBEAM_OAUTH_TIMEOUT', 15))
"""How many seconds to wait for the OAuth login request to complete."""

PAGE_LOAD_TIMEOUT = int(os.environ.get('IBEAM_PAGE_LOAD_TIMEOUT', 15))
"""How many seconds to wait for the login page to load."""

ERROR_SCREENSHOTS = to_bool(os.environ.get('IBEAM_ERROR_SCREENSHOTS', False))
"""Whether to save login page screenshots on error."""

MAX_FAILED_AUTH = int(os.environ.get('IBEAM_MAX_FAILED_AUTH', 5))
"""Maximum number of failed authentication attempts."""

MIN_PRESUBMIT_BUFFER = int(os.environ.get('IBEAM_MIN_PRESUBMIT_BUFFER', 5))
"""Minimum number of seconds to wait before hitting the submit button"""

MAX_PRESUBMIT_BUFFER = int(os.environ.get('IBEAM_MAX_PRESUBMIT_BUFFER', 30))
"""Maximum number of seconds to wait before hitting the submit button"""

MAX_IMMEDIATE_ATTEMPTS = int(os.environ.get('IBEAM_MAX_IMMEDIATE_ATTEMPTS', 10))
"""Maximum number of immediate retries upon detecting an error message."""

IBKEY_PROMO_EL_CLASS = os.environ.get('IBEAM_IBKEY_PROMO_EL_CLASS', 'CLASS_NAME@@ibkey-promo-skip')
"""HTML element class containing the ibkey promo skip button."""

AUTHENTICATION_STRATEGY = os.environ.get('IBEAM_AUTHENTICATION_STRATEGY', 'B')
"""The authentication strategy used by IBeam."""

MAX_STATUS_CHECK_RETRIES = int(os.environ.get('IBEAM_MAX_STATUS_CHECK_RETRIES', 120))
"""How many times to reattempt the status check."""

MAX_REAUTHENTICATE_RETRIES = int(os.environ.get('IBEAM_MAX_REAUTHENTICATE_RETRIES', 3))
"""How many times to reattempt the reauthentication before restarting the Gateway."""

UI_SCALING = float(os.environ.get('IBEAM_UI_SCALING', 1))
"""The resolution UI scaling to be used by the browser."""

########### TWO-FACTOR AUTHENTICATION ###########

TWO_FA_EL_ID = os.environ.get('IBEAM_TWO_FA_EL_ID', 'ID@@twofactbase')
"""HTML element check for if Gateway will require 2FA code authentication."""

TWO_FA_NOTIFICATION_EL = os.environ.get('IBEAM_TWO_FA_NOTIFICATION_EL', 'CLASS_NAME@@login-step-notification')
"""HTML element check for if Gateway will require 2FA notification authentication."""

TWO_FA_INPUT_EL_ID = os.environ.get('IBEAM_TWO_FA_INPUT_EL_ID', 'ID@@xyz-field-bronze-response')
"""HTML element to input 2FA code into"""

TWO_FA_HANDLER = os.environ.get('IBEAM_TWO_FA_HANDLER', None)
"""Which 2FA handler should be used to acquire the code."""

STRICT_TWO_FA_CODE = to_bool(os.environ.get('IBEAM_STRICT_TWO_FA_CODE', True))
"""Whether to ensure only 2FA code made of 6 digits can be used."""

TWO_FA_SELECT_EL_ID = os.environ.get('IBEAM_TWO_FA_SELECT_EL_ID', 'ID@@xyz-field-bronze-response')
"""HTML element check for if Gateway requires to select the 2FA method."""

TWO_FA_SELECT_TARGET = os.environ.get('IBEAM_TWO_FA_SELECT_TARGET', 'IB Key')
"""Option that is to be chosen in the 2FA select dropdown"""

CUSTOM_TWO_FA_HANDLER = os.environ.get('IBEAM_CUSTOM_TWO_FA_HANDLER', 'custom_two_fa_handler.CustomTwoFaHandler')
"""Fully qualified path of the custom 2FA handler in the inputs directory."""

all_variables = {key: value for key, value in vars().items() if (not key.startswith("__") and key.isupper() and key != 'UNDEFINED')}



import logging
import os
import re
import time
import urllib.parse
from pathlib import Path
from typing import Optional

from selenium.common.exceptions import ElementClickInterceptedException

from ibeam.src.login.driver import release_chrome_driver, save_screenshot, DriverFactory
from ibeam.src.utils.selenium_utils import any_of
from ibeam.src.two_fa_handlers.two_fa_handler import TwoFaHandler

from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

_LOGGER = logging.getLogger('ibeam.' + Path(__file__).stem)

_GOOG_QR_CODE_CLASS = os.environ.get('IBEAM_GOOG_QR_CODE_CLASS', 'bigger-qr-code')
"""HTML element indicating web messages needs authorization."""

_GOOG_QR_CODE_DATA = os.environ.get('IBEAM_GOOG_QR_CODE_DATA', 'qr-code')
"""HTML data attribute with the qr code."""

_GOOG_AUTH_REMEMBER_CLASS = os.environ.get('IBEAM_GOOG_AUTH_REMEMBER_CLASS', 'local-storage-checkbox')
"""HTML element to remember web messages device pairing."""

_GOOG_MESSAGES_LIST_CLASS = os.environ.get('IBEAM_GOOG_MESSAGES_LIST_CLASS', '.text-content.unread .snippet-text')
"""HTML element indicating web messages has loaded."""

_GOOG_2FA_HEADING = os.environ.get('IBEAM_GOOG_2FA_HEADING', 'Your requested authentication code')
"""HTML element text indicating 2fa message received."""

_GOOG_MESSAGE_CLICK_RETRIES = int(os.environ.get('IBEAM_GOOG_MESSAGE_CLICK_RETRIES', 5))
"""How many times to try marking the message as read."""


class GoogleMessagesTwoFaHandler(TwoFaHandler):

    def __init__(self, driver_factory:DriverFactory, *args, **kwargs):
        self.driver_factory = driver_factory
        super().__init__(*args, **kwargs)

    def get_two_fa_code(self, _) -> Optional[str]:
        code_two_fa = None

        driver_2fa = self.driver_factory.new_driver(name='google_msg', incognito=False)
        if driver_2fa is None:
            return None

        try:
            driver_2fa.get('https://messages.google.com/web')

            sms_auth_present = EC.presence_of_element_located((By.CLASS_NAME, _GOOG_QR_CODE_CLASS))
            sms_code_present = EC.text_to_be_present_in_element((By.CSS_SELECTOR, _GOOG_MESSAGES_LIST_CLASS),
                                                                _GOOG_2FA_HEADING)

            WebDriverWait(driver_2fa, 240).until(any_of(sms_auth_present, sms_code_present))

            sms_auth_el = driver_2fa.find_elements(By.CLASS_NAME, _GOOG_QR_CODE_CLASS)

            if sms_auth_el:
                driver_2fa.find_element(By.CLASS_NAME, _GOOG_AUTH_REMEMBER_CLASS).click()

                data = urllib.parse.quote(sms_auth_el[0].get_attribute('data-' + _GOOG_QR_CODE_DATA))

                _LOGGER.info(
                    'Web messages is not authenticated. Open this URL to pair web messages with your android phone:')
                _LOGGER.info(
                    f'http://api.qrserver.com/v1/create-qr-code/?color=000000&bgcolor=FFFFFF&qzone=1&margin=0&size=400x400&ecc=L&data={data}')

                WebDriverWait(driver_2fa, 120).until(sms_code_present)

            sms_list_el = driver_2fa.find_elements(By.CSS_SELECTOR, _GOOG_MESSAGES_LIST_CLASS)

            if not sms_list_el:
                _LOGGER.error('Timeout or authentication error while loading sms messages.')
                save_screenshot(driver_2fa, self.outputs_dir, postfix='__google_2fa')
            else:
                _LOGGER.info(f'First SMS found: "{sms_list_el[0].text}"')

                code_two_fa = re.search(r'(\d+)', sms_list_el[0].text).group(1)

                _LOGGER.info('Waiting for SMS message to be visible')
                WebDriverWait(driver_2fa, 30).until(EC.visibility_of(sms_list_el[0]))

                clicked_ok = False
                for i in range(_GOOG_MESSAGE_CLICK_RETRIES):
                    try:
                        sms_list_el[0].click()  # mark message as read
                        clicked_ok = True
                        _LOGGER.info('SMS message marked as read')
                        break
                    except ElementClickInterceptedException as e:
                        if isinstance(e, ElementClickInterceptedException) \
                                and 'Other element would receive the click' in str(e):
                            _LOGGER.warning(f'Failed marking SMS message as read due to obstructing elements')
                        else:
                            _LOGGER.exception(f'Exception while marking SMS message as read: {e}')

                        save_screenshot(driver_2fa, self.outputs_dir, postfix='__google_2fa')

                        _LOGGER.info(f'Retrying clicking SMS message {_GOOG_MESSAGE_CLICK_RETRIES - i - 1} more times.')
                        time.sleep(2)

                if not clicked_ok:
                    _LOGGER.warning('Failed all attempts to mark SMS message as read')

                time.sleep(2)  # wait for click to mark message as read

        except:
            save_screenshot(driver_2fa, self.outputs_dir, '__google-msg')
            raise
        finally:
            _LOGGER.info(f'Cleaning up the resources. Google MSG Driver: {driver_2fa}')
            release_chrome_driver(driver_2fa)

        return code_two_fa

    def __str__(self):
        return f"GoogleMessagesTwoFaHandler(driver_path={self.driver_factory.driver_path})"
